#!/usr/bin/perl -w
#
# relDate.pl v2.3
# Last modified September 8, 2011 14:05
# (c) Olaf R.P. Bininda-Emonds
#
# Purpose:
#	Calculates the relative branch lengths on a gene tree that can be used to
#	help date a (super)tree
#
# Input:
#	1) A nexus- or phylip-formatted tree file containing the supertree tree
#	   description
#	2) A nexus- or phylip-formatted tree file containing the descriptions of one
#	   or more gene trees (with branch lengths)
#
# Output:
#	Up to two tab-delimited text files (relDate_ancDate.txt and/or
#	relDate_daughterDate.txt) containing relative branch lengths for all the
#	supertree nodes for each gene tree.
#
#	Each file is designed to be imported into an Excel worksheet, where the
#	formulas reference either relative (default) or named cells. (To produce
#	the latter, insert a blank column in column B and select columns A and B.
#	Select Insert > Name > Create ... from the menus and create labels in the
#	left column only.) Cells will have one of three values: 1) blank if the gene
#	tree does not provide information for the corresponding supertree node,
#	2) FALSE if information is provided, but the reference cell does not have
#	any information, or 3) the relative date. Note that the values in the named
#	reference cells (column B) must default to FALSE if no information is
#	present.
#
# Usage: relDate.pl -g<filename> -t<filename> [-d<a|d|b>] [-m] [-n] [-r] [-v]
#		 [-h]
#   options: -d<a|d|b> = calculate dates relative to ancestral (a -- default) or
#						 daughter nodes only (d) or both individually (b)
#			 -g<filename> = tree file containing gene trees in NEXUS or phylip
#							formats
#			 -m = indicate that genes trees contain multiple specimens of
#				  Genus_species in form Genus_species_x
#			 -n = have formulas reference named cells in Excel spreadsheet
#				  (default = relative cells in column B)
#			 -r = assume trees to be rooted unless otherwise specified (default
#				  = unrooted)
#            -t<filename> = tree file containing supertree in NEXUS or phylip
#							formats
#			 -v = verbose output
#            -h = print this message and quit

use strict;

# Set user-input defaults and associated parameters
	# Supertree parsing
		my $superFile;
		my (%STnodeList, $numSTclade, @STtaxa, %taxonST);
	
	# Gene tree parsing
		my $geneFile;
		my (%mismatchFlag, %skipTree);
		my $multSpecimens = 0;

	# treeRead module variables		
		my $block = 1;
			my $treeBlock;
		my (%treeName, %treeDescription, %rootingStatus);
		my %prunedTree;
		my $unrootedFlag = 0;
		my $defaultRooting = "[&U]";	# Default is for trees to be unrooted
		my $SupportFlag = 0;
		my @allTaxa;
		my $treeNum;
		
	# Branch length variables
#		my $outFile = "relDate_results.txt";
			my $ancDateFile = "relDate_ancDate.txt";
			my $dauDateFile = "relDate_daughterDate.txt";
#		my %relHeight;
		my (%ancHeight, %dauHeight);
		my $useNodes = "ancestral";

	# Output variables
		my $refCell = "relative";	# Options are "named" and "relative"
		my %excelCell;
	
	# Global variables
		my $debug = 0;
		my $verbose = 0;
		my $version = "2.3";

# Process switches
	for (my $i = 0; $i <= $#ARGV; $i++)
		{
		if ($ARGV[$i] eq "-h")
			{
			print "Usage: relDate.pl -g<filename> -t<filename> [-d<a|d|b>] [-m] [-n] [-r] [-v]\n";
			print "       [-h]\n";
			print "Version: $version\n";
			print "Options: -d<a|d|b> = calculate dates relative to ancestral (a -- default) or\n";
			print "                     daughter nodes only (d) or both individually (b)\n";
			print "         -g<filename> = tree file containing gene trees in NEXUS or phylip\n";
			print "                        formats\n";
			print "         -m = indicate that genes trees contain multiple specimens of\n";
			print "              Genus_species in form Genus_species_x\n";
			print "         -n = have formulas reference named cells in Excel spreadsheet\n";
			print "              (default = relative cells in column B)\n";
			print "         -r = assume trees to be rooted unless otherwise specified\n";
			print "              (default = unrooted)\n";
			print "         -t<filename> = tree file containing supertree in NEXUS or phylip\n";
			print "                        formats\n";
			print "         -v = verbose output\n";
			print "         -h = print this message and quit\n";
			exit(0);
			}
		elsif ($ARGV[$i] eq "-da")
			{
			$useNodes = "ancestral";
			}
		elsif ($ARGV[$i] eq "-dd")
			{
			$useNodes = "daughter";
			}
		elsif ($ARGV[$i] eq "-db")
			{
			$useNodes = "both";
			}
		elsif ($ARGV[$i] =~ /^-g(.+)/)
			{
			$geneFile = $1;
			die "ERROR: Treefile $geneFile does not exist.\n" unless (-e $geneFile);
			}
		elsif ($ARGV[$i] eq "-m")
			{
			$multSpecimens = 1;
			}
		elsif ($ARGV[$i] eq "-n")
			{
			$refCell = "named";
			}
		elsif ($ARGV[$i] eq "-r")
			{
			$defaultRooting = "[&R]";
			}
		elsif ($ARGV[$i] =~ /^-t(.+)/)
			{
			$superFile = $1;
			die "ERROR: Treefile $superFile does not exist.\n" unless (-e $superFile);
			(my $rootFile = $superFile) =~ s/\.\w+$//;
				$ancDateFile = $rootFile . "_ancDate.txt";
				$dauDateFile = $rootFile . "_daughterDate.txt";
			}
		elsif ($ARGV[$i] eq "-v")
			{
			$verbose = 1;
			}
		elsif ($ARGV[$i] eq "-x")
			{
			$debug = $verbose = 1;
			}
		else
			{
			print "Don't understand argument: $ARGV[$i]\n";
			print "Usage: relDate.pl -g<filename> -t<filename> [-d<a|d|b>] [-m] [-n] [-r] [-v] [-h]\n";
			print "Version: $version\n";
			exit(1); 
			}
		}
	die "ERROR: Must specify name of treefile containing supertree topology.\n" if (not defined $superFile);
	die "ERROR: Must specify name of treefile containing gene tree(s).\n" if (not defined $geneFile);

# Read in supertree and store information
	print "Reading in and parsing supertree description from file $superFile ...\n";
	my $superZero = time; 
	$block = 0;
	my @superTrees = treeRead($superFile);
		my $superTreeID = shift(@superTrees);
		my $superTree = $treeDescription{$superTreeID};
			$superTree =~ s/:\-?\d+\.?\d*e?\-?\d*//ig;	# Remove any branch length information
			$numSTclade = ($superTree =~ tr/(//);
				die "ERROR: Supertree does not have enough informative nodes ($numSTclade clades)\n" if ($numSTclade < 2);
			print "\nSupertree description:\t$superTree\n" if ($verbose);

	# Parse supertree to get clade memberships and sizes
		my ($cladeID, %ParentClade, %CladeDescrip, %CladeSupport, %STcladeSize, %open, %start, %stop, %present);
		my ($position, $cladeNum, $CurrentClade) = (0, 0, 0);
		my $OldCurrent;
		while ($position < length($superTree))
			{
			my $offset = 1;
			if (substr($superTree, $position, 1) eq "(")		#New clade; go one level deeper
				{
				$cladeNum++;
					$cladeID = "STnode" . $cladeNum;
					$excelCell{$cladeID} = "B" . ($cladeNum + 2);
				$ParentClade{$cladeID} = $CurrentClade;
					$ParentClade{$cladeID} = "STnode0" if ($cladeID eq "STnode1");
				$CurrentClade = $cladeID;
				$open{$CurrentClade} = 1;
				$start{$cladeID} = $position;
				}
			elsif (substr($superTree, $position, 1) eq ")")	#Clade finished; clean up and go up one level
				{
				$open{$CurrentClade} = 0;
				$OldCurrent = $CurrentClade;
				$CurrentClade = $ParentClade{$OldCurrent};
				$stop{$OldCurrent} = $position;
				$CladeDescrip{$OldCurrent} = substr($superTree, $start{$OldCurrent}, $stop{$OldCurrent}-$start{$OldCurrent}+1);
				$CladeDescrip{$OldCurrent} =~ s/:\-?\d+\.?\d*e?\-?\d*//ig;	# Remove branch length information
				}
			elsif (substr($superTree, $position, 1) eq ",")
				{
				# Do nothing
				}
			else	# Determine membership of clade
				{
				until (substr($superTree, $position+$offset, 1) eq "(" or substr($superTree, $position+$offset, 1) eq ")" or substr($superTree, $position+$offset, 1) eq "," or substr($superTree, $position+$offset, 1) eq ":")
					{
					$offset++;
					}
				my $taxon = substr($superTree, $position, $offset);
					push @STtaxa, $taxon;
					$taxonST{$taxon} = 1;
				for (my $i = 1; $i <= $cladeNum; $i++)	# Assign all open clades to taxon and increment size of open clades
					{
					my $clade = "STnode" . $i;
					if ($open{$clade})
						{
						push @{ $STnodeList{$taxon} }, $clade;
						$STcladeSize{$clade}++;
						}
					}
				$position += $offset-1;
				}
			$position++;
			}
	
		@STtaxa = sort @STtaxa;
	
		if ($debug)
			{
			print "\nTaxon membership and size of supertree clades:\n";
				for (my $i = 1; $i <= $numSTclade; $i++)
					{
					my $clade = "STnode" . $i;
					print "\t$clade ($STcladeSize{$clade}): $CladeDescrip{$clade}\n" if ($verbose);
					} 
			print "\nSupertree clade membership of taxa:\n";
				foreach my $taxon (@STtaxa)
					{
					print "\t$taxon:";
					foreach my $node (@{ $STnodeList{$taxon} })
						{
						print "\t$node";
						}
					print "\n";
					}
			}
	printf "\tTime taken: %s seconds\n", time - $superZero;

# Read in gene trees and store
	print "\nReading in gene trees ...\n";
	$block = 1;
	my @geneTrees = treeRead($geneFile);

	# Parse gene tree descriptions and get relative branch lengths
		my $treeCount;
		foreach my $geneTreeID (@geneTrees)	# Get structure of each tree
			{
#			my (%ancHeight, %dauHeight);
			my (%ancCladeHeight, %dauCladeHeight);
			
			my $geneZero = time;
			
			my $geneTree = $treeDescription{$geneTreeID};
			$geneTree =~ s/:\d+\.?\d*e?\-?\d*$//i;	# Get rid of trailing branch length, which is uninformative and only causes problems when parsing branch lengths
			my (%ParentClade, %daughterClade);
			my ($cladeID, %CladeDescrip, %open, %start, %stop);
			my ($taxon, %geneNodeMember, @GTtaxa, @mismatchList);
			my (%CladeSupport, %nodeDepth, %cladeDepth, %cladeHeight);
				# Set variables for root
					$nodeDepth{GTnode0} = 0;
					$cladeDepth{GTnode0} = 0;
					$CladeSupport{GTnode1} = 0;			
			my (%gene2ST);

			$treeCount++;
				print "\tProcessing tree $treeName{$geneTreeID} ...\n";
				if ($verbose)
					{
					print "\t$geneTree\n";
					print "\t\tTree is ";
						print "un" if ($rootingStatus{$geneTreeID} eq "[&U]");
					print "rooted\n";
					}
			# Do general check on tree structure
				# Check for mismatched taxa
					my $treeTaxa = $geneTree;
						$treeTaxa =~ s/:\-?\d+\.?\d*e?\-?\d*//ig;
						$treeTaxa =~ s/\(//g;
						$treeTaxa =~ s/\)//g;
						foreach my $taxonCheck (split(/,/, $treeTaxa))
							{
							(my $rootName = $taxonCheck) =~ s/_\d+$//;
							unless ($taxonST{$taxonCheck} or $taxonST{$rootName})	# List any taxa not present on supertree
								{
								push @mismatchList, $taxonCheck;
								}
							}
						if (@mismatchList)
							{
							$skipTree{$treeCount} = 1;
							open (MISMATCH, ">>mismatch.txt") or die "Cannot append mismatched taxa to file mismatch.txt.\n";
								@mismatchList = sort (@mismatchList);
								print "\tERROR: Tree contains taxa not on supertree; list written to file mismatch.txt\n";
								print MISMATCH "$treeName{$geneTreeID}:";
								foreach my $booboo (@mismatchList)
									{
									print MISMATCH "\t$booboo";
									}
								print MISMATCH "\n";
							close MISMATCH;
							}
				# Check tree size
					my $totalClade = ($geneTree =~ tr/(//);
						if ($totalClade < 2)
							{
							print "\tERROR: Tree has only one informative node\n";
							$skipTree{$treeCount} = 1;
							}
				# Check for branch lengths
					unless ($geneTree =~ m/:\d+/)
						{
						print "\tERROR: Tree does not contain any branch lengths\n";
						$skipTree{$treeCount} = 1;
						}
				next if ($skipTree{$treeCount});
				
			print "\t\tParsing structure ...\n";
			my ($position, $cladeNum, $CurrentClade) = (0, 0, 0);
			my $OldCurrent;
			while ($position < length($geneTree))
				{
				my $offset = 1;
				if (substr($geneTree, $position, 1) eq "(")		# New clade; go one level deeper
					{
					$cladeNum++;
						$cladeID = "GTnode" . $cladeNum;
					$ParentClade{$cladeID} = $CurrentClade;
						$ParentClade{$cladeID} = "GTnode0" if ($cladeID eq "GTnode1");
					push @ {$daughterClade{$CurrentClade} }, $cladeID;	# Assign new clade as daughter clade to parent clade
					$CurrentClade = $cladeID;
					$nodeDepth{$CurrentClade} = $nodeDepth{$ParentClade{$CurrentClade}} + 1;
					$open{$CurrentClade} = 1;
					$start{$cladeID} = $position;
					}
				elsif (substr($geneTree, $position, 1) eq ")")	# Clade finished; clean up and go up one level
					{
					$open{$CurrentClade} = 0;
					$OldCurrent = $CurrentClade;
					$CurrentClade = $ParentClade{$OldCurrent};
					$stop{$OldCurrent} = $position;

					# Get clade description
						$CladeDescrip{$OldCurrent} = substr($geneTree, $start{$OldCurrent}, $stop{$OldCurrent}-$start{$OldCurrent}+1);
						$CladeDescrip{$OldCurrent} =~ s/:\-?\d+\.?\d*e?\-?\d*//ig;	# Remove branch length information
					print "\t\t\tClade ID: $OldCurrent\n" if ($debug);
					print "\t\t\t\t$CladeDescrip{$OldCurrent}\n" if ($debug);
					
					# Determine corresponding supertree clade
						my %count;
						print "\t\t\t\tTaxon membership:\n" if ($debug);
						foreach my $geneTaxon (@{ $geneNodeMember{$OldCurrent} })	# Check each species of node in gene tree
							{
							print "\t\t\t\t\t$geneTaxon\n" if ($debug);
							foreach my $STnode (@{ $STnodeList{$geneTaxon} })	# Find what nodes that species belongs to in supertree
								{
								$count{$STnode}++;
								}
							}
						# Find supertree node of smallest size for which all species in gene tree are present; this is the least inclusive node on supertree containing all taxa from the node on the gene tree
							print "\t\t\t\tFinding equivalent ST node:\n" if ($debug);
							if ($gene2ST{$OldCurrent})	# Clade is mutliple specimens of a single species; no equivalent
								{
								print "\t\t\t\t\tnone\n" if ($debug);
								}
							else
								{
								my $maxSize = scalar(@STtaxa) + 1;	# Initialize counter to one larger than number of taxa on ST
								for (my $STnodeNum = 1; $STnodeNum <= $numSTclade; $STnodeNum++)
									{
									my $STnode = "STnode" . $STnodeNum;
									# Quick checks to avoid extra processing
										next unless ($count{$STnode});
										next if (scalar(@{ $geneNodeMember{$OldCurrent} }) > $STcladeSize{$STnode});	# Gene tree clade is more inclusive than ST clade
										next if ($STcladeSize{$STnode} > $maxSize);
									print "\t\t\t\t\tST node ID: $STnode ($STcladeSize{$STnode})\n" if ($debug);
									if ($count{$STnode} == scalar(@{ $geneNodeMember{$OldCurrent} }) and $STcladeSize{$STnode} < $maxSize)
		
										{
										$gene2ST{$OldCurrent} = $STnode;
										$maxSize = $STcladeSize{$STnode};
										print "\t\t\t\t\t\tCurrent equivalent\n" if ($debug);
										}
									}
								}
					# Determine absolute and relative heights from tips
						my $sumHeight = 0;
						print "\t\t\t\tDaughter clades:\n" if ($debug);
						foreach my $daughter (@ {$daughterClade{$OldCurrent} })
							{
							print "\t\t\t\t\t$daughter: $CladeSupport{$daughter}\n" if ($debug);
							$sumHeight += $CladeSupport{$daughter};							
							$sumHeight += $cladeHeight{$daughter};							
							}
						print "\t\t\t\t\tSum height: $sumHeight\n" if ($debug);
						$cladeHeight{$OldCurrent} = sprintf "%.5f", $sumHeight / scalar(@ {$daughterClade{$OldCurrent} });
							print "\t\t\t\t\tClade height: $cladeHeight{$OldCurrent}\n" if ($debug);
					}
				elsif (substr($geneTree, $position, 1) eq ":")	# Store branch lengths for node or taxon
					{
					until (substr($geneTree, $position+$offset, 1) eq "(" or substr($geneTree, $position+$offset, 1) eq ")" or substr($geneTree, $position+$offset, 1) eq "," or ($position+$offset) > length($geneTree))
						{
						$offset++;
						}
					my $branchLength = substr($geneTree, $position+1, $offset-1);
						print "\t\t\t\tBranch length: $branchLength\n" if ($debug);
						if (substr($geneTree, $position-1, 1) eq ")")	# Branch length pertains to a clade
							{
							$CladeSupport{$OldCurrent} = substr($geneTree, $position+1, $offset-1);
							}
						else	# Branch length pertains to a terminal taxon
							{
							$CladeSupport{$taxon} = $branchLength;
							}
					$position += $offset-1;
					}
				elsif (substr($geneTree, $position, 1) eq ",")
					{
					# Do nothing
					}
				else	# Determine membership of clade
					{
					until (substr($geneTree, $position+$offset, 1) eq "(" or substr($geneTree, $position+$offset, 1) eq ")" or substr($geneTree, $position+$offset, 1) eq "," or substr($geneTree, $position+$offset, 1) eq ":")
						{
						$offset++;
						}
					$taxon = substr($geneTree, $position, $offset);
						push @ { $daughterClade{$CurrentClade}}, $taxon;
						push @GTtaxa, $taxon;
						$ParentClade{$taxon} = $CurrentClade;
						$nodeDepth{$taxon} = $nodeDepth{$CurrentClade} + 1;
						$cladeHeight{$taxon} = 0;
					print "\t\t\tTaxon: $taxon\n" if ($debug);

					my $realTaxon = $taxon;	# Need to keep taxon name as read intact for branch length association
					if ($realTaxon =~ m/_\d+$/ and $multSpecimens)	# Taxon of form Genus_species_x; repeated specimen of Genus_species
						{
						$gene2ST{$cladeID} = "none";	# No supertree equivalent possible for node
						if ($realTaxon =~m/_1$/)
							{
							$realTaxon =~ s/_1$//;	# Get Genus_species to be stored; will work even if Genus_sepcies_1 is listed merely as Genus_species (will not enter loop at next most inclusive level)
							}
						else
							{
							undef $realTaxon;	# To prevent storing same Genus_species multiple times; only species name will be stored
							}
						}
					if ($realTaxon)
						{
#						unless ($taxonST{$realTaxon})	# List any taxa not present on supertree
#							{
#							push @mismatchList, $realTaxon;
#							}
						for (my $i = 1; $i <= $cladeNum; $i++)	# Assign taxon to any open clades
							{
							my $clade = "GTnode" . $i;
							if ($open{$clade})
								{
								push @{ $geneNodeMember{$clade} }, $realTaxon;
#								$geneNodeMember{substr($geneTree, $position, $offset)}{$clade} = 1;
								}
							}
						}
					$position += $offset-1;
					}
				$position++;
				}

			@GTtaxa = sort(@GTtaxa);

			# Get relative heights of all nodes; also set clade depths from tips for all nodes and taxa
				print "\t\tDetermining relative branch lengths relative to supertree ...\n";
				for (my $i = 1; $i <= $totalClade; $i++)
					{
					my $iClade = "GTnode" . $i;
					$cladeDepth{$iClade} = sprintf "%.5f", $CladeSupport{$iClade} + $cladeDepth{$ParentClade{$iClade}};

					# Determine relative height
						unless ($skipTree{$treeCount})
							{
							if ($rootingStatus{$geneTreeID} eq "[&U]" and ($ParentClade{$iClade} eq "GTnode0" or $ParentClade{$iClade} eq "GTnode1"))	# Tree is unrooted and clade descends from or is "root"; ancestral branch length is meaningless
								{
								$ancHeight{$treeCount}{$gene2ST{$iClade}} = $dauHeight{$treeCount}{$gene2ST{$iClade}} = "undefined";
								$ancCladeHeight{$iClade} = $dauCladeHeight{$iClade} = "undefined";
								}
							elsif ($cladeHeight{$ParentClade{$iClade}} or ($cladeHeight{$iClade} and $iClade =~ /^GTnode/))
								{
								# Calculate ancestral heights
									$ancHeight{$treeCount}{$gene2ST{$iClade}} = "undefined";	# Defaults to undefined
									if ($cladeHeight{$ParentClade{$iClade}})
										{
										if ($refCell eq "relative")
											{
											$ancHeight{$treeCount}{$gene2ST{$iClade}} = "=if(count($excelCell{$gene2ST{$ParentClade{$iClade}}}), $cladeHeight{$iClade} / ($cladeHeight{$iClade} + $CladeSupport{$iClade}) * $excelCell{$gene2ST{$ParentClade{$iClade}}}, FALSE())";
											$ancHeight{$treeCount}{$gene2ST{$iClade}} = "=if(count($excelCell{$gene2ST{$ParentClade{$iClade}}}), 1 * $excelCell{$gene2ST{$ParentClade{$iClade}}}, FALSE())" if ($cladeHeight{$iClade} + $CladeSupport{$iClade} == 0);	# Safety in case denominator sums to zero
											}
										else
											{
											$ancHeight{$treeCount}{$gene2ST{$iClade}} = "=if(count($gene2ST{$ParentClade{$iClade}}), $cladeHeight{$iClade} / ($cladeHeight{$iClade} + $CladeSupport{$iClade}) * $gene2ST{$ParentClade{$iClade}}, FALSE())";
											$ancHeight{$treeCount}{$gene2ST{$iClade}} = "=if(count($gene2ST{$ParentClade{$iClade}}), 1 * $gene2ST{$ParentClade{$iClade}}, FALSE())" if ($cladeHeight{$iClade} + $CladeSupport{$iClade} == 0);	# Safety in case denominator sums to zero
											}
										print "\t\t\t$iClade\tancestral: $ancHeight{$treeCount}{$gene2ST{$iClade}}\n" if ($debug);
										}
									$ancCladeHeight{$iClade} = $ancHeight{$treeCount}{$gene2ST{$iClade}};

								# Calculate daughter heights
									$dauHeight{$treeCount}{$gene2ST{$iClade}} = "undefined";	# Defaults to undefined
									if ($cladeHeight{$iClade})
										{
										my @daughterHeights;
										my $countStatement = "count(";
										foreach my $daughter (@ { $daughterClade{$iClade}})
											{
											# Skip daughters that are 1) terminals, 2) do not have an equivalent on the supertree, or 3) have a branch length equal to or exceeding height of parent node (up to third decimal place)
												next unless ($daughter =~ /^GTnode/);
												next if ($gene2ST{$daughter} eq "none");
												next if (sprintf("%.3f", $CladeSupport{$daughter} - $cladeHeight{$iClade}) >= 0);
#													next if ($CladeSupport{$daughter} >= $cladeHeight{$iClade});

											if ($refCell eq "relative")
												{
												push @daughterHeights, "($cladeHeight{$iClade} + $CladeSupport{$daughter}) / $cladeHeight{$iClade} * $excelCell{$gene2ST{$daughter}}";
												$countStatement .= $excelCell{$gene2ST{$daughter}} . ",";
												}
											else
												{
												push @daughterHeights, "($cladeHeight{$iClade} + $CladeSupport{$daughter}) / $cladeHeight{$iClade} * $gene2ST{$daughter}";
												$countStatement .= $gene2ST{$daughter} . ",";
												}
											}
											$countStatement =~ s/,$/\)/;
											
										if (scalar @daughterHeights >= 1)
											{
											$dauHeight{$treeCount}{$gene2ST{$iClade}} = join(', ', @daughterHeights);
											if (scalar @daughterHeights > 1)
												{
												$dauHeight{$treeCount}{$gene2ST{$iClade}} = "average(" . $dauHeight{$treeCount}{$gene2ST{$iClade}} . ")";
												}
											$dauHeight{$treeCount}{$gene2ST{$iClade}} = "=if(" . $countStatement . ", " . $dauHeight{$treeCount}{$gene2ST{$iClade}} . ", FALSE())";
											}
											
										print "\t\t\t$iClade\tdaughter: $dauHeight{$treeCount}{$gene2ST{$iClade}}\n" if ($debug);
										}
									$dauCladeHeight{$iClade} = $dauHeight{$treeCount}{$gene2ST{$iClade}};

								# Calculate relative height
# 									$relHeight{$treeCount}{$gene2ST{$iClade}} = "undefined";	# Defaults to undefined
# 									if ($useNodes eq "ancestral")
# 										{
# 										$relHeight{$treeCount}{$gene2ST{$iClade}} = "= $ancHeight{$treeCount}{$gene2ST{$iClade}}" unless ($ancHeight{$treeCount}{$gene2ST{$iClade}} eq "undefined");
# 										}
# 									elsif ($useNodes eq "daughter")
# 										{
# 										$relHeight{$treeCount}{$gene2ST{$iClade}} = "= $dauHeight{$treeCount}{$gene2ST{$iClade}}" unless ($dauHeight{$treeCount}{$gene2ST{$iClade}} eq "undefined");
# 										}
# 									else	# ($useNodes eq "both")
# 										{
# 										if ($ancHeight{$treeCount}{$gene2ST{$iClade}} ne "undefined" and $dauHeight{$treeCount}{$gene2ST{$iClade}} ne "undefined")
# 											{
# 											$relHeight{$treeCount}{$gene2ST{$iClade}} = "= average($ancHeight{$treeCount}{$gene2ST{$iClade}}, $dauHeight{$treeCount}{$gene2ST{$iClade}})";
# 											}
# 										if ($ancHeight{$treeCount}{$gene2ST{$iClade}} ne "undefined" and $dauHeight{$treeCount}{$gene2ST{$iClade}} eq "undefined")
# 											{
# 											$relHeight{$treeCount}{$gene2ST{$iClade}} = "= $ancHeight{$treeCount}{$gene2ST{$iClade}}";
# 											}
# 										if ($ancHeight{$treeCount}{$gene2ST{$iClade}} eq "undefined" and $dauHeight{$treeCount}{$gene2ST{$iClade}} ne "undefined")
# 											{
# 											$relHeight{$treeCount}{$gene2ST{$iClade}} = "= $dauHeight{$treeCount}{$gene2ST{$iClade}}";
# 											}
# 										}
# 									$relCladeHeight{$iClade} = $relHeight{$treeCount}{$gene2ST{$iClade}};
# 									print "\t\t\t$iClade\trelative: $relHeight{$treeCount}{$gene2ST{$iClade}}\n" if ($debug);
								}
							else
								{
								$ancHeight{$treeCount}{$gene2ST{$iClade}} = $dauHeight{$treeCount}{$gene2ST{$iClade}} = "undefined";
#								$ancHeight{$treeCount}{$gene2ST{$iClade}} = $dauHeight{$treeCount}{$gene2ST{$iClade}} = $relHeight{$treeCount}{$gene2ST{$iClade}} = "undefined";
								$ancCladeHeight{$iClade} = $dauCladeHeight{$iClade} = "undefined";
								print "\t\t\t$iClade\tundefined\n" if ($debug);
								}
							}
					}
				foreach my $species (@GTtaxa)
					{
					$cladeDepth{$species} = sprintf "%.5f", $CladeSupport{$species} + $cladeDepth{$ParentClade{$species}};
					}

			# Print out node / taxon statistics
				if ($verbose)
					{
					print "\t\tClade statistics:\n";
					for (my $i = 1; $i <= $totalClade; $i++)
						{
						my $iClade = "GTnode" . $i;
						print "\t\t\tClade ID: $iClade\n";
						print "\t\t\t\tDescription: $CladeDescrip{$iClade}\n";
						print "\t\t\t\t\tCorresponding supertree node: $gene2ST{$iClade}\n";
						print "\t\t\t\tParent clade ID: ";
							if ($i == 1)
								{
								print "none (basal clade)\n";
								}
							else
								{
								print "$ParentClade{$iClade}\n";
								print "$CladeDescrip{$ParentClade{$iClade}}\n" if ($debug);
								}
						print "\t\t\t\tDaughter clade IDs: ";
							if (@ { $daughterClade{$iClade}})
								{
								foreach my $daughter (@ { $daughterClade{$iClade}})
									{
									print "\t$daughter";
									}
								}
							else
								{
								print "\tnone";
								}
							print "\n";
						print "\t\t\t\tBranch length: ";
							if ($CladeSupport{$iClade} and $i > 1)
								{
								print "$CladeSupport{$iClade}\n";
								}
							else
								{
								print "none\n";
								}
						print "\t\t\t\t\tNode depth (from root): $nodeDepth{$iClade}\n";
						print "\t\t\t\t\tClade depth (from root): $cladeDepth{$iClade}\n";
						print "\t\t\t\t\tClade height (from tips): $cladeHeight{$iClade}\n" if ($cladeHeight{$iClade});
						print "\t\t\t\t\tAncestral height: $ancCladeHeight{$iClade}\n";
						print "\t\t\t\t\tDaughter height: $dauCladeHeight{$iClade}\n";
#						print "\t\t\t\t\tRelative height: $relCladeHeight{$iClade}\n";
						}

					foreach my $taxon (@GTtaxa)
						{
						print "\t\t\tTaxon name: $taxon\n";
						print "\t\t\t\tParent clade: $CladeDescrip{$ParentClade{$taxon}}\t(Clade ID: $ParentClade{$taxon})\n";
						print "\t\t\t\tBranch length: $CladeSupport{$taxon}\n";
						print "\t\t\t\t\tNode depth (from root): $nodeDepth{$taxon}\n";
						print "\t\t\t\t\tClade depth (from root): $cladeDepth{$taxon}\n";
						print "\t\t\t\t\tClade height (from tips): $cladeHeight{$taxon}\n";
						}
					}
			printf "\t\tTime taken: %s seconds\n", time - $geneZero;
			}

# Print results!
	print "\nPrinting results ...\n";
		printExcel($ancDateFile) if ($useNodes eq "ancestral" or $useNodes eq "both");
		printExcel($dauDateFile) if ($useNodes eq "daughter" or $useNodes eq "both");
			
exit(0);

### Subroutines used in script

sub treeRead	# Subroutine to read in tree descriptions and derive taxon names; accounts for any translation tables
	{
	my $treeFile = shift;

	setLineBreak($treeFile);
	open (TREES, "<$treeFile") or die "Cannot open nexus-formatted treefile $treeFile\n";
		my (@allTrees, $rawTreeDescrip, $treeDescrip, $treeFlag, $treeCount, $rooting);
		my (@taxonList, %taxonPresent);
		my (%TTable, $translation, $translateFlag,$maxTaxonID, $transText);
		my $commentFlag = 0;
		$treeBlock = 0;
		
		while (<TREES>)
			{
			chomp;
			$_ =~ s/^\s+//;		# Strip (any) leading white space
			$_ =~ s/\s+$//;		# Strip (any) trailing white space
			
			# Block any commented lines
				if ($_ =~ /^\[/)
					{
					$commentFlag += ($_ =~ tr/\[+//);
					}
				if ($_ =~ /\]\s?$/)
					{
					$_ =~ s/\s+$//;
					$commentFlag -= ($_ =~ tr/\]+//);
					next;
					}
				next if ($commentFlag > 0);
				
			# Process various nexus commands
				if ($_ =~ /^begin trees;/i)		# Clear variables for each tree block
					{
					undef %TTable;
					$translation = $translateFlag = $treeFlag = $treeCount = 0;
					$treeBlock++;
					}
				if ($_ =~ /^translate/i)	# Read translation table of taxon names (if any)
					{
					print "Reading translation table ...\n" if ($verbose);
					$translateFlag = $translation = 1;
					$transText = "";
					$maxTaxonID = 0;
					}
				if ($translateFlag)
					{
					unless ($_ =~ /^translate/i)
						{
						$transText .= $_;
						if ($_ =~ /;/)	# End of translate command; parse
							{
							$translateFlag = 0;
							$transText =~ s/;$//;	#Strip trailing semi-colon
							my @transEntries = split(/,/, $transText);
							foreach my $entry (@transEntries)
								{
								$entry =~ s/^\s+//;		#Strip (any) leading white space
								my @transEntry = split(/\s/,$entry);
									my $taxonID = shift(@transEntry);
									my $taxonName = join(' ', @transEntry);

								$maxTaxonID = $taxonID if ($taxonID > $maxTaxonID);
								$taxonName =~ s/\s+/_/g;
									$taxonName = "'" . $taxonName . "'" if ($taxonName =~ /\W/);
									$taxonName =~ s/^\'+/\'/;
									$taxonName =~ s/\'+$/\'/;
								$TTable{$taxonID} = $taxonName;
								$taxonPresent{$taxonName}++;
								push @taxonList, $taxonName;
								}
							}
						}
					}
				if (($_ =~ /tree/i and $_ =~ /=/) or $_ =~ /^\(/)
					{
					unless ($_ =~ /^>/)		#Ignore (any) commented lines
						{
						$rawTreeDescrip = $treeDescrip = "";
						unless ($block == 0 and $treeBlock > 1)
							{
							$treeFlag = 1;
							}
						}
					}
	
				if ($treeFlag)
					{
					$rawTreeDescrip .= $_;
					if ($_ =~ /:/)
						{
						$SupportFlag = 1;
						}
					if ($_ =~ /;/)	# End of tree description; parse
						{
						$treeCount++;
						print "\nTree block: $treeBlock\tTree number: $treeCount\n" if ($verbose);
						print "\tTranslating taxon names ...\n" if ($verbose and $translation);
						# Determine if trees are rooted or not (default for no information is not rooted)
							if ($rawTreeDescrip =~ /\[&R\]/)
								{
								$rooting = "[&R]";
								}
							elsif ($rawTreeDescrip =~ /\[&U\]/)
								{
								$rooting = "[&U]";
								$unrootedFlag = 1;
								}
							else
								{
								$rooting = $defaultRooting;
								$unrootedFlag = 1 if ($defaultRooting eq "[&U]");
								}
						# Remove any embedded comments
							if ($rawTreeDescrip =~ /\[/)
								{
								for (my $position = 0; $position < length($rawTreeDescrip); $position++)
									{
									if (substr($rawTreeDescrip, $position, 1) eq "[")
										{
										$position++ until (substr($rawTreeDescrip, $position, 1) eq "]");
										}
									else
										{
										$treeDescrip .= substr($rawTreeDescrip, $position, 1);
										}
									}
								}
							else
								{
								$treeDescrip = $rawTreeDescrip;
								}
						$treeFlag = 0;
						my $position = 0;
						until (substr($treeDescrip, $position, 1) eq "(")		# Avoid anything before tree description proper
							{
							$position++;
							}
						my $tree = substr($treeDescrip, $position);
							$tree =~ s/;\s*$//;
							$tree =~ s/\s+//g;
							
						my $name;
						if ($treeDescrip =~ /=/)
							{
							($name) = (split /=/, $treeDescrip, 0);		# Get name of tree, if any
							$name =~ s/\s*\w*tree\s+//i;
							$name =~ s/\s*\*\s*//i;
							$name =~ s/\s+$//g;
							$name =~ s/-/_/g;
							$name =~ s/\s+/_/g;
							$name =~ s/^'//;
							$name =~ s/'$//;
							if ($name =~ /\W/)
								{
								$name = "'" . $name . "'";
								}
#							$treeNames{$treeCount} = $name;
							}
						my $openCount = ($tree =~ tr/(//);
						my $closedCount = ($tree =~ tr/)//);
						if ($openCount == $closedCount)		# Tree is balanced; process
							{
							$position = length($tree) - 1;
							while ($position >= 0)
								{
								my $IDlength = 1;
								if (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")		#Tree structure; skip to next character
									{
									$position--;
									}
								else	# Read in taxonID
									{
									until (substr($tree, $position, 1) eq "(" or substr($tree, $position, 1) eq ")" or substr($tree, $position, 1) eq "," or substr($tree, $position, 1) eq ":")
										{
										$position--;
										$IDlength++;
										}
									unless (substr($tree, $position, 1) eq ":" or substr($tree, $position, 1) eq ")")		# Avoid reading branch lengths or node labels as taxon names
										{
										my $readName = substr($tree, $position+1, $IDlength-1);
										if ($translation and not defined $taxonPresent{$readName})	# Translate taxon names; latter condition is a safety for BioPerl convert_trees.pl module which adds, but doesn't use translation block
											{
											if ($TTable{$readName})
												{
												substr($tree, $position+1, $IDlength-1) = $TTable{$readName};		# Translate name
												unless ($taxonPresent{$TTable{$readName}})
													{
													push @taxonList, $TTable{$readName};
													$taxonPresent{$TTable{$readName}} = 1;
													}
												}
											else
												{
												print "\t\tUnidentified taxon ".$readName." in tree #".$treeCount." of tree block #".$treeBlock.". Replaced with xxx.\n";
												substr($tree, $position+1, $IDlength-1) = "xxx";		# Translate name
												unless ($taxonPresent{"xxx"})
													{
													push @taxonList, "xxx";
													$taxonPresent{"xxx"} = 1;
													}
												}
											}
										else	# Enter name as read from tree, unless already present
											{
											$taxonPresent{$readName}++;
											push @taxonList, $readName if ($taxonPresent{$readName} == 1);
											}
										}
									}
								}
								
							$treeNum++;
								my $treeID = "treeID_$treeNum";
								$name = "readTree_$treeNum" if (not defined $name);
							
							push @allTrees, $treeID;
								$treeDescription{$treeID} = $tree;
								$rootingStatus{$treeID} = $rooting;
								$treeName{$treeID} = $name;
							print "\tTree description:\n\t\t$tree\n" if ($verbose);
							}
						else	#Tree is unbalanced; reject
							{
							print "\tTree #".$treeCount." of tree block #".$treeBlock." is unbalanced. Not processed.\n";
							}
						}
					}
			}
	close TREES;
	
	# Process taxon list
		@allTaxa = sort(duplStrip(@taxonList));
	
	return @allTrees;
	}

sub duplStrip	# Subroutine to strip duplicate entries from an array
	{
	my @duplArray = @_;
	my %dupl;		#Remove duplicate taxa
		my @stripArray = grep(!$dupl{$_}++, @duplArray);
		@stripArray = sort {$a cmp $b} @stripArray;
	return @stripArray;
	}

sub printExcel
	{
	my $outFile = shift;
	my %relHeight;
		%relHeight = %ancHeight if ($outFile =~ /ancDate/);
		%relHeight = %dauHeight if ($outFile =~ /daughterDate/);
	
	open (EXCEL, ">$outFile") or die "Cannot open results file, $outFile.\n";
		# Print header row
			print EXCEL "Data file derived from $superFile and $geneFile using relDate v$version on ".localtime()."\n";
			print EXCEL "ST node";
				print EXCEL "\tDate" if ($refCell eq "relative");
				print EXCEL "\t$treeName{$_}" foreach (@geneTrees);
			print EXCEL "\n";
		for (my $STclade = 1; $STclade <= $numSTclade; $STclade++)
			{
			my $STnode = "STnode" . $STclade;
			print EXCEL "$STnode";
				print EXCEL "\t" if ($refCell eq "relative");
			print "$STnode" if ($debug);
			for (my $tree = 1; $tree <= $treeCount; $tree++)
				{
				if ($skipTree{$tree})
					{
					print EXCEL "\t";
					}
				elsif ($relHeight{$tree}{$STnode})
					{
					if ($relHeight{$tree}{$STnode} eq "undefined")
						{
						print EXCEL "\t";
						}
					else
						{
						print EXCEL "\t$relHeight{$tree}{$STnode}";
						}
					print "\t$relHeight{$tree}{$STnode}" if ($debug);
					}
				else
					{
					print EXCEL "\t";
					print "\t$STnode\tn/a" if ($debug);
					}
				}
			print EXCEL "\n";
			print "\n" if ($debug);
			}
	close EXCEL;
	print "\tResults printed to file, $outFile.\n";
	}

sub setLineBreak	# Check line breaks of input files and set input record separator accordingly
	{
	my $inFile = shift;
	$/ ="\n";
	open (IN, "<$inFile") or die "Cannot open $inFile to check form of line breaks.\n";
		while (<IN>)
			{
			if ($_ =~ /\r\n/)
				{
				print "\tDOS line breaks detected ...\n" if ($verbose);
				$/ ="\r\n";
				last;
				}
			elsif ($_ =~ /\r/)
				{
				print "\tMac line breaks detected ...\n" if ($verbose);
				$/ ="\r";
				last;
				}
			else
				{
				print "\tUnix line breaks detected ...\n" if ($verbose);
				$/ ="\n";
				last;
				}
			}
	close IN;
	}

# Version history
#
#	v2.3 (September 8, 2011)
#		- relative dates can now be determined using both ancestral and daughter
#		  nodes (albeit separately)
#		- major bug fixes
#			- fixed bug causing daughter-derived dates to be (massive)
#			  overestimates
#			- fixed bug causing terminal taxa to sometimes be associated with
#			  wrong parent node; bug could impinge on all dates throughout tree,
#			  but more likely and more severely for nodes closer to the tips
#		- fixed calculation of relative date formulas to account for missing
#		  data in reference cell
#		- fixed Excel equations so as to use semi-colons and not commas in "if"
#		  statements
#		- now have to explicitly tell script that multiple specimens of form
#		  Genus_species_x exist in gene trees using -m switch (caused problems
#		  when real taxa were of this form)
#		- added GNU GPL statement
#		- reading of trees now uses treeRead module
#		- checks that necessary input files exist before proceeding
#		- adds version information to output file
#		- minor bug fixes
#
#	v2.2 (June 6, 2005)
#		- relative dates can now be determined using ancestral or daughter nodes
#		  (user-specified)
#		- nodes on tree now have a unique identifier (and not just numbered) in
#		  case one or more terminal taxa
#		  have purely numeric names
#		- minor bug fixes
#
#	v2.1 (February 7, 2005)
#		- correction for potential division by zero error
#		- automatically detect or allow user to specify rooting status of trees
#		  and handle accordingly
#		- minor improvements to interface
#		- minor bug fixes
#
#	v2.0 (July 28, 2004)
#		- initial release (v1.x was only used in-house)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# A copy of the GNU General Public License is available at
# http://www.gnu.org/copyleft/gpl.html or by writing to
# the Free Software Foundation, Inc., 51 Franklin Street,
# Fifth Floor, Boston, MA, 02110-1301, USA.
